/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package projectcompiler;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.sql.SQLException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.ResourceBundle;
import java.util.logging.Level;
import java.util.logging.Logger;
import javafx.application.Platform;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.fxml.Initializable;
import javafx.scene.Node;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.control.Alert;
import javafx.scene.control.Label;
import javafx.scene.layout.AnchorPane;
import javafx.scene.layout.Region;
import javafx.scene.layout.StackPane;
import javafx.stage.FileChooser;
import javafx.stage.Modality;
import javafx.stage.Stage;
import javafx.stage.StageStyle;
import javafx.stage.WindowEvent;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;

/**
 *
 * @author Richard Orilla
 */
public class MainInterfaceController implements Initializable {
    
    private Label label;
    @FXML
    private StackPane machineCode;
    @FXML
    private StackPane userCode;
    @FXML
    private StackPane assemblerCode;
    private boolean withComments = true;
    
    String lastCompile = "";
    Node parent;
    @FXML
    private AnchorPane anchorPane;
    
    @Override
    public void initialize(URL url, ResourceBundle rb) {

    }    

    @FXML
    private void openFile(ActionEvent event) {
        FileChooser fileChooser = new FileChooser();
        fileChooser.setTitle("Open Resource File"); 
        File fileSelected = fileChooser.showOpenDialog(machineCode.getScene().getWindow());
        FileInputStream inputStream = null;
        try {
            inputStream = new FileInputStream(fileSelected);
        } catch (FileNotFoundException ex) {
            Logger.getLogger(MainInterfaceController.class.getName()).log(Level.SEVERE, null, ex);
        }
        try {
            String everything = IOUtils.toString(inputStream, "UTF-8");
            ProjectCompiler.c1.replaceText(everything);
            inputStream.close();
        } catch (IOException ex) {
            Logger.getLogger(MainInterfaceController.class.getName()).log(Level.SEVERE, null, ex);
        } finally {
            
        }
    }

    @FXML
    private void closeProgram(ActionEvent event) {
        Platform.exit();
    }

    @FXML
    private void compileCode(ActionEvent event) throws IOException {
        if (!lastCompile.equals(ProjectCompiler.c1.getText())) {
            Process p = null;

            writeToFile("lang.hll", ProjectCompiler.c1.getText());
            lastCompile = ProjectCompiler.c1.getText();
            String input = "lang.hll";
            String output = "asm.arc";
            boolean hllHasError = false;
            String hllErrors = "";
            String arcErrors = "";
            boolean arcHasError = false;
            
            try {
                p = Runtime.getRuntime().exec("hll\\compiler.exe " + input + " " + output);
                p.waitFor();
                BufferedReader reader = new BufferedReader(new InputStreamReader(p.getInputStream()));
                String line=reader.readLine();

                while (line != null) {    
                    System.out.println(line);
                    if (!line.isEmpty()) {
                        hllHasError = true;
                        hllErrors += line + "\n";
                    }
                    line = reader.readLine();
                }
            } catch (IOException ex) {
                Logger.getLogger(ProjectCompiler.class.getName()).log(Level.SEVERE, null, ex);
            } catch (InterruptedException ex) {
                Logger.getLogger(ProjectCompiler.class.getName()).log(Level.SEVERE, null, ex);
            }
            if (hllHasError) {
                Alert msg = new Alert(Alert.AlertType.ERROR);
                msg.initStyle(StageStyle.UTILITY);
                msg.initModality(Modality.WINDOW_MODAL);
                msg.initOwner(anchorPane.getScene().getWindow());
                msg.getDialogPane().setMinHeight(Region.USE_PREF_SIZE);
                msg.setTitle("Compilation Error");
                msg.setHeaderText("HLL Compiler encountered an error");
                msg.setContentText(hllErrors);
                msg.showAndWait();
                ProjectCompiler.c2.replaceText(hllErrors);
                ProjectCompiler.c3.replaceText("Could not generate binary because arc.exe recieved an empty code");
                return;
            }
            ProjectCompiler.c2.replaceText(readFromFile("asm.arc"));
            input = "asm.arc";
            output = "code.mac";
            String comment = withComments ? "" : " -nocomments";

            try {
                p = Runtime.getRuntime().exec("arc\\compiler.exe " + input + " " + output + comment);
                p.waitFor();
                BufferedReader reader = new BufferedReader(new InputStreamReader(p.getInputStream()));
                String line=reader.readLine();

                while (line != null) {    
                    System.out.println(line);
                    if (line.contains("unsuccessful")) {
                        arcHasError = true;
                    }
                    else {
                        arcErrors += line + "\n";
                    }
                    line = reader.readLine();
                }
            } catch (IOException ex) {
                Logger.getLogger(ProjectCompiler.class.getName()).log(Level.SEVERE, null, ex);
            } catch (InterruptedException ex) {
                Logger.getLogger(ProjectCompiler.class.getName()).log(Level.SEVERE, null, ex);
            }
            if (arcHasError) {
                Alert msg = new Alert(Alert.AlertType.ERROR);
                msg.initStyle(StageStyle.UTILITY);
                msg.initModality(Modality.WINDOW_MODAL);
                msg.initOwner(anchorPane.getScene().getWindow());
                msg.getDialogPane().setMinHeight(Region.USE_PREF_SIZE);
                msg.setTitle("Compilation Error");
                msg.setHeaderText("ARC Compiler encountered an error");
                msg.setContentText(arcErrors);
                msg.showAndWait();
                ProjectCompiler.c3.replaceText(arcErrors);
                return;
            }
            ProjectCompiler.c3.replaceText(readFromFile("code.mac"));
            Alert msg = new Alert(Alert.AlertType.INFORMATION);
            msg.initStyle(StageStyle.UTILITY);
            msg.initModality(Modality.WINDOW_MODAL);
            msg.initOwner(anchorPane.getScene().getWindow());
            msg.getDialogPane().setMinHeight(Region.USE_PREF_SIZE);
            msg.setTitle("Compilation Successful");
            msg.setHeaderText("Compilation Success");
            msg.setContentText("The program has successfully compiled it to the corresponding machine code. " +
                    "Several files are created based on its compilation. These are:\n\n" +
                    "1. lang.hll - For the high level language\n" + 
                    "2. asm.arc - The corresponding assembler code\n" + 
                    "3. code.mac - The corresponding machine code\n\n" +
                    "To remove the comments from the corresponding machine code, please uncheck Allow Comments option in the option menu."
            );
            msg.showAndWait();
        }
        else {
            Alert msg = new Alert(Alert.AlertType.WARNING);
            msg.initStyle(StageStyle.UTILITY);
            msg.initModality(Modality.WINDOW_MODAL);
            msg.initOwner(anchorPane.getScene().getWindow());
            msg.getDialogPane().setMinHeight(Region.USE_PREF_SIZE);
            msg.setTitle("Compilation Aborted");
            msg.setContentText("The system has detected no changes with the underlying source code.");
            msg.showAndWait();
        }
    }

    @FXML
    private void showAbout(ActionEvent event) {
        FXMLLoader load = new FXMLLoader(getClass().getResource("AboutInterface.fxml"));
        Parent root = null;
        Scene scene = null;
        Stage stage = new Stage();
        try {
            root = load.load();
        } catch (IOException ex) {
            Logger.getLogger(MainInterfaceController.class.getName()).log(Level.SEVERE, null, ex);
        }
        scene = new Scene(root);
        stage.setScene(scene);
        stage.initModality(Modality.WINDOW_MODAL);
        stage.initOwner(anchorPane.getScene().getWindow());
        stage.initStyle(StageStyle.UTILITY);
        stage.setTitle("About this program");
        stage.setResizable(false);
        stage.show();
    }
    
    private void writeToFile(String fileName, String str) throws IOException {
        BufferedWriter writer = new BufferedWriter(new FileWriter(fileName, false));
        writer.write(str.replace("\n", "\r\n"));
        writer.close();
    }
    
    private String readFromFile(String fileName)throws IOException {
        File file = new File(fileName);
        return FileUtils.readFileToString(file, "UTF-8");
    }

    @FXML
    private void spawnHLL(ActionEvent event) {
    }

    @FXML
    private void spawnARC(ActionEvent event) {
        FXMLLoader load = new FXMLLoader(getClass().getResource("ARCSettings.fxml"));
        Parent root = null;
        Scene scene = null;
        Stage stage = new Stage();
        try {
            root = load.load();
        } catch (IOException ex) {
            Logger.getLogger(MainInterfaceController.class.getName()).log(Level.SEVERE, null, ex);
        }
        scene = new Scene(root);
        stage.setScene(scene);
        stage.initModality(Modality.WINDOW_MODAL);
        stage.initOwner(anchorPane.getScene().getWindow());
        stage.initStyle(StageStyle.UTILITY);
        stage.setTitle("About this program");
        stage.setResizable(false);
        stage.show();
        ARCSettingsController controller = load.getController();
        controller.setParent(this);
    }
    
    public boolean getARCComments() {
        return withComments;
    }
    
    public void setARCComments(boolean v) {
        withComments = v;
    }
    
}
