/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package projectcompiler;

import java.time.Duration;
import java.util.Collection;
import java.util.Collections;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javafx.application.Application;
import javafx.fxml.FXMLLoader;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.image.Image;
import javafx.scene.layout.AnchorPane;
import javafx.scene.layout.StackPane;
import javafx.stage.Stage;

import org.fxmisc.flowless.VirtualizedScrollPane;
import org.fxmisc.richtext.CodeArea;
import org.fxmisc.richtext.LineNumberFactory;
import org.fxmisc.richtext.model.StyleSpans;
import org.fxmisc.richtext.model.StyleSpansBuilder;
import org.reactfx.Subscription;

/**
 *
 * @author Richard Orilla
 */
public class ProjectCompiler extends Application {
    
    private static final String[] KEYWORDS = new String[] {
            "for", "do", "end", "if", "turn", "execute", "->"
    };
    
    private static final String[] KEYWORDS2 = new String[] {
            "dcl", "asg", "cmp", "add", "sub", "mul", "jmp", "fwd", "lft", "rgt", "rdm", "wtm", "set", "sns"
    };
    
    private static final String[] KEYWORDS3 = new String[] {
            "00000"
    };
    
    private static final String[] KEYWORDS4 = new String[] {
            "01000"
    };

    private static final String KEYWORD_PATTERN = "\\b(" + String.join("|", KEYWORDS) + ")\\b";
    private static final String KEYWORD_PATTERN2 = "\\b(" + String.join("|", KEYWORDS2) + ")\\b";
    private static final String KEYWORD_PATTERN3 = "\\b(" + String.join("|", KEYWORDS3) + ")\\b";
    private static final String KEYWORD_PATTERN4 = "\\b(" + String.join("|", KEYWORDS4) + ")\\b";
    
    private static final String PAREN_PATTERN = "\\(|\\)";
    private static final String BRACE_PATTERN = "\\{|\\}";
    private static final String BRACKET_PATTERN = "\\[|\\]";
    private static final String SEMICOLON_PATTERN = "\\;";
    private static final String STRING_PATTERN = "\"([^\"\\\\]|\\\\.)*\"";
    private static final String COMMENT_PATTERN = "//[^\n]*" + "|" + "/\\*(.|\\R)*?\\*/";
    private static final String COMMENT_PATTERN2 = "//[^\n]*" + "|" + "/\\*(.|\\R)*?\\*/";
    private static final String COMMENT_PATTERN3 = "//[^\n]*" + "|" + "/\\*(.|\\R)*?\\*/";

    private static final Pattern PATTERN = Pattern.compile(
                "(?<KEYWORD>" + KEYWORD_PATTERN + ")"
                + "|(?<PAREN>" + PAREN_PATTERN + ")"
                + "|(?<BRACE>" + BRACE_PATTERN + ")"
                + "|(?<BRACKET>" + BRACKET_PATTERN + ")"
                + "|(?<SEMICOLON>" + SEMICOLON_PATTERN + ")"
                + "|(?<STRING>" + STRING_PATTERN + ")"
                + "|(?<COMMENT>" + COMMENT_PATTERN + ")"
    );
    
    private static final Pattern PATTERN2 = Pattern.compile(
                "(?<KEYWORD>" + KEYWORD_PATTERN2 + ")"
                + "|(?<COMMENT>" + COMMENT_PATTERN2 + ")"
    );
    
    private static final Pattern PATTERN3 = Pattern.compile(
                "(?<KEYWORD>" + KEYWORD_PATTERN3 + ")"
                + "|(?<COMMENT>" + COMMENT_PATTERN3 + ")"
    );
    
    private static final String sampleCode1 = String.join("\r\n", new String[] {
            "a = 0",
            "for (a -> 10) do",
            "  b = 5 + 2 - (1 * 3)",
            "  execute forward",
            "  if (a > 5) then",
            "     turn left",
            "  else",
            "     turn right",
            "  end if",
            "end",
            "c = b"
    });
    
    public static CodeArea c1, c2, c3;
    
    @Override
    public void start(Stage stage) throws Exception {
        CodeArea codeArea1 = new CodeArea();
        codeArea1.setParagraphGraphicFactory(LineNumberFactory.get(codeArea1));
        Subscription cleanupWhenNoLongerNeedIt1 = codeArea1
                .multiPlainChanges()
                .successionEnds(Duration.ofMillis(500))
                .subscribe(ignore -> codeArea1.setStyleSpans(0, computeHighlighting(codeArea1.getText())));
        codeArea1.replaceText(0, 0, sampleCode1);
        
        CodeArea codeArea2 = new CodeArea();
        codeArea2.setParagraphGraphicFactory(LineNumberFactory.get(codeArea2));
        Subscription cleanupWhenNoLongerNeedIt2 = codeArea2
                .multiPlainChanges()
                .successionEnds(Duration.ofMillis(500))
                .subscribe(ignore -> codeArea2.setStyleSpans(0, computeHighlighting2(codeArea2.getText())));
        
        CodeArea codeArea3 = new CodeArea();
        codeArea3.setParagraphGraphicFactory(LineNumberFactory.get(codeArea3));
        Subscription cleanupWhenNoLongerNeedIt3 = codeArea3
                .multiPlainChanges()
                .successionEnds(Duration.ofMillis(500))
                .subscribe(ignore -> codeArea3.setStyleSpans(0, computeHighlighting3(codeArea3.getText())));
        
        Parent root = FXMLLoader.load(getClass().getResource("MainInterface.fxml"));
        AnchorPane ref = (AnchorPane)root;
        Scene scene = new Scene(ref);
        //Scene scene = new Scene(new StackPane(new VirtualizedScrollPane<>(codeArea)), 600, 400);
        stage.setScene(scene);
        stage.show();
        stage.setTitle("Richard Orilla - CSC617M for Sir Solomon See");
        stage.getIcons().add(new javafx.scene.image.Image("file:icon.png"));
        StackPane userCode = (StackPane) ref.lookup("#userCode");
        StackPane assemblerCode = (StackPane) ref.lookup("#assemblerCode");
        StackPane machineCode = (StackPane) ref.lookup("#machineCode");
        userCode.getChildren().add(new StackPane(new VirtualizedScrollPane<>(codeArea1)));
        assemblerCode.getChildren().add(new StackPane(new VirtualizedScrollPane<>(codeArea2)));
        machineCode.getChildren().add(new StackPane(new VirtualizedScrollPane<>(codeArea3)));
        c1 = codeArea1; c2 = codeArea2; c3 = codeArea3;
        userCode.getStylesheets().add(ProjectCompiler.class.getResource("custom-keywords.css").toExternalForm());
        assemblerCode.getStylesheets().add(ProjectCompiler.class.getResource("arc-keywords.css").toExternalForm());
        machineCode.getStylesheets().add(ProjectCompiler.class.getResource("mac-keywords.css").toExternalForm());
        c2.setEditable(false);
        c3.setEditable(false);
    }

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        launch(args);
    }
    
    private static StyleSpans<Collection<String>> computeHighlighting(String text) {
        Matcher matcher = PATTERN.matcher(text);
        int lastKwEnd = 0;
        StyleSpansBuilder<Collection<String>> spansBuilder
                = new StyleSpansBuilder<>();
        while(matcher.find()) {
            String styleClass =
                    matcher.group("KEYWORD") != null ? "keyword" :
                    matcher.group("PAREN") != null ? "paren" :
                    matcher.group("BRACE") != null ? "brace" :
                    matcher.group("BRACKET") != null ? "bracket" :
                    matcher.group("SEMICOLON") != null ? "semicolon" :
                    matcher.group("STRING") != null ? "string" :
                    matcher.group("COMMENT") != null ? "comment" :
                    null; /* never happens */ assert styleClass != null;
            spansBuilder.add(Collections.emptyList(), matcher.start() - lastKwEnd);
            spansBuilder.add(Collections.singleton(styleClass), matcher.end() - matcher.start());
            lastKwEnd = matcher.end();
        }
        spansBuilder.add(Collections.emptyList(), text.length() - lastKwEnd);
        return spansBuilder.create();
    }
    
    private static StyleSpans<Collection<String>> computeHighlighting2(String text) {
        Matcher matcher = PATTERN2.matcher(text);
        int lastKwEnd = 0;
        StyleSpansBuilder<Collection<String>> spansBuilder
                = new StyleSpansBuilder<>();
        while(matcher.find()) {
            String styleClass =
                    matcher.group("KEYWORD") != null ? "keyword" :
                    matcher.group("PAREN") != null ? "paren" :
                    matcher.group("BRACE") != null ? "brace" :
                    matcher.group("BRACKET") != null ? "bracket" :
                    matcher.group("SEMICOLON") != null ? "semicolon" :
                    matcher.group("STRING") != null ? "string" :
                    matcher.group("COMMENT") != null ? "comment" :
                    null; /* never happens */ assert styleClass != null;
            spansBuilder.add(Collections.emptyList(), matcher.start() - lastKwEnd);
            spansBuilder.add(Collections.singleton(styleClass), matcher.end() - matcher.start());
            lastKwEnd = matcher.end();
        }
        spansBuilder.add(Collections.emptyList(), text.length() - lastKwEnd);
        return spansBuilder.create();
    }
    
    private static StyleSpans<Collection<String>> computeHighlighting3(String text) {
        Matcher matcher = PATTERN3.matcher(text);
        int lastKwEnd = 0;
        StyleSpansBuilder<Collection<String>> spansBuilder
                = new StyleSpansBuilder<>();
        while(matcher.find()) {
            String styleClass =
                    matcher.group("KEYWORD") != null ? "keyword" :
                    matcher.group("PAREN") != null ? "paren" :
                    matcher.group("BRACE") != null ? "brace" :
                    matcher.group("BRACKET") != null ? "bracket" :
                    matcher.group("SEMICOLON") != null ? "semicolon" :
                    matcher.group("STRING") != null ? "string" :
                    matcher.group("COMMENT") != null ? "comment" :
                    null; /* never happens */ assert styleClass != null;
            spansBuilder.add(Collections.emptyList(), matcher.start() - lastKwEnd);
            spansBuilder.add(Collections.singleton(styleClass), matcher.end() - matcher.start());
            lastKwEnd = matcher.end();
        }
        spansBuilder.add(Collections.emptyList(), text.length() - lastKwEnd);
        return spansBuilder.create();
    }
    
}
