﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace highlevelanguage
{
    class Program
    {

        static int loopCount = 1;
        static int ifCount = 1;
        static Dictionary<string, string> variableList;

        static void Main(string[] args)
        {
            GParser parser = new GParser();
            parser.Setup("maingrammar.egt");
            if (args.Length == 0)
            {
                string[] g = { "input.txt", "output.txt" };
                args = g;
            }
            
            if (args.Length >= 2)
            {
                System.IO.StreamReader file = new System.IO.StreamReader(args[0]);
                System.IO.StreamWriter output = new System.IO.StreamWriter(args[1]);
                string line = "";
                int lineNum = 0;
                string wholeLine = "";
                while ((line = file.ReadLine()) != null)
                {
                    if (line != "")
                    {
                        wholeLine += " " + line + "\r\n";
                        lineNum += 1;
                    }
                }
                //Console.WriteLine(wholeLine);
                if (parser.Parse(new StringReader(wholeLine)))
                {
                    //Console.WriteLine("Success");
                    List<Line> lines = parser.lines;
                    List<ExprResultToken> lineresult = new List<ExprResultToken>();
                    List<string> errorList = new List<string>();
                    variableList = new Dictionary<string, string>();

                    bool isUnderLoop = false;
                    bool isUnderCondition = false;
                    for (int i = 0; i < lines.Count; i++)
                    {
                        ExprResultToken[] currentLine = new ExprResultToken[0];
                        try
                        {
                            switch (lines[i].tokenlist[0].ToString())
                            {
                                case "execute":
                                    currentLine = execute(lines[i]);
                                    break;
                                case "turn":
                                    currentLine = turn(lines[i]);
                                    break;
                                case "sense":
                                    currentLine = sense(lines[i]);
                                    break;
                                case "for":
                                    //Console.WriteLine("for");
                                    isUnderLoop = true;
                                    currentLine = startLoop(lines[i]);
                                    break;
                                case "end":
                                    //Console.WriteLine("end");
                                    isUnderLoop = false;
                                    currentLine = endLoop(lines[i]);
                                    loopCount += 1;
                                    break;
                                case "if":
                                    isUnderCondition = true;
                                    currentLine = ifCMD(lines[i]);
                                    break;
                                case "else":
                                    currentLine = elseCMD(lines[i]);
                                    break;
                                case "end if":
                                    isUnderCondition = false;
                                    currentLine = endIfCMD(lines[i]);
                                    break;
                                default:
                                    //Console.WriteLine("Variable Assignment");
                                    try
                                    {
                                        if (!variableExist(lines[i].tokenlist[0].ToString()))
                                        {
                                            variableList.Add(lines[i].tokenlist[0].ToString(), lines[i].tokenlist[0].ToString());
                                            currentLine = initVariable(lines[i]);
                                        }
                                        else
                                        {
                                            //Console.WriteLine(lines[i].tokenlist[0].ToString());
                                            currentLine = setVariable(lines[i]);
                                        }
                                        
                                    }
                                    catch (Exception ex)
                                    {
                                        
                                    }
                                    break;
                            }
                        }
                        catch (Exception error)
                        {
                            errorList.Add("Line #" + (i+1).ToString()  + " " + error.Message);
                        }

                        if (!isUnderCondition && !isUnderLoop)
                        {
                            lineresult.AddRange(currentLine);
                        }
                        else if (isUnderLoop)
                        {
                            lineresult.AddRange(currentLine);
                        }
                    }
                    int maxValue = 0;
                    foreach (ExprResultToken cline in lineresult)
                    {
                        if (cline.container[0] == "add" || cline.container[0] == "sub" || cline.container[0] == "mul" ||
                            cline.container[0] == "dcl" || cline.container[0] == "asg")
                        {
                            for (int i = 1; i < cline.container.Count; i++)
                            {
                                if(cline.container[i].ToString()[0] == '@')
                                {
                                    int val = int.Parse(cline.container[i].ToString().Substring(1));
                                    if (val > maxValue)
                                    {
                                        maxValue = val;
                                    }
                                }
                            }
                        }
                    }
                    ExprResultToken[] currentLine2 = prepend(maxValue);
                    if (maxValue > 0)
                    {
                        lineresult.InsertRange(0, currentLine2);
                    }
                    if (errorList.Count == 0)
                    {
                        foreach (ExprResultToken cline in lineresult)
                        {
                            if (cline.container[0] == "add" || cline.container[0] == "sub" || cline.container[0] == "mul" ||
                                cline.container[0] == "dcl" || cline.container[0] == "asg" || cline.container[0] == "sns")
                            {
                                for (int i = 0; i < cline.container.Count; i++)
                                {

                                    output.Write(cline.container[i]);
                                    if (i + 1 != cline.container.Count && i != 0)
                                    {
                                        output.Write(", ");
                                    }
                                    else if (i == 0)
                                    {
                                        output.Write(" ");
                                    }
                                    else
                                    {
                                        output.Write("\r\n");
                                    }
                                }
                            }
                            else
                            {
                                output.WriteLine(cline);
                            }

                        }
                    }
                    else
                    {
                        foreach(string error in errorList)
                        {
                            Console.WriteLine(error);
                        }
                    }
                }
                else
                {
                    Console.WriteLine(parser.FailMessage);
                }
                file.Close();
                output.Close();
            }
            //Console.ReadLine();
        }

        static bool variableExist(string name)
        {
            try
            {
                string obtain = variableList[name];
                return true;
            }
            catch(Exception ex)
            {
                return false;
            }
        }

        static bool isNumber(string value)
        {
            try
            {
                int.Parse(value);
                return true;
            }
            catch (Exception ex)
            {
                return false;
            }
        }

        static ExprResultToken[] prepend(int val)
        {
            ExprResultToken[] result = new ExprResultToken[val];
            for (int i = 0; i < val; i++)
            {
                ExprToken declToken = new ExprToken("raw", "dcl");
                ExprToken varToken = new ExprToken("raw", "@" + (i+1));
                ExprToken valToken = new ExprToken("raw", "0");
                //
                ExprResultToken declLine = new ExprResultToken(new ExprToken[] { declToken, varToken, valToken });
                result[i] = declLine;
            }
            return result;
        }

        static ExprResultToken [] initVariable(Line line)
        {
            List<ExprToken> cline = convertToExprTree(line.tokenlist.GetRange(2, line.tokenlist.Count - 2));
            List<ExprResultToken> init;
            if (cline.Count > 1)
            {
                init = exprBreakdown(ref cline);
                try
                {
                    ExprResultToken lastLine = init[init.Count - 1];
                    ExprToken finalReference = lastLine.container[lastLine.container.Count - 1];
                    ExprToken declareToken = new ExprToken("raw", "dcl");
                    ExprToken variableName = new ExprToken("raw", line.tokenlist[0].value);
                    ExprResultToken declaration = new ExprResultToken(new ExprToken[] { declareToken, variableName, finalReference });
                    init.Add(declaration);
                    /*foreach (ExprResultToken l in init)
                    {
                        Console.WriteLine(l);
                    }*/
                }
                catch (Exception ex)
                {

                }
                return init.ToArray();
            }
            else
            {
                init = new List<ExprResultToken>();
                ExprToken declareToken = new ExprToken("raw", "dcl");
                ExprToken variableName = new ExprToken("raw", line.tokenlist[0].value);
                ExprToken finalReference = new ExprToken("raw", line.tokenlist[2].value);
                ExprResultToken declaration = new ExprResultToken(new ExprToken[] { declareToken, variableName, finalReference });
                init.Add(declaration);
                /*foreach (ExprResultToken l in init)
                {
                    Console.WriteLine(l);
                }*/
            }
            return init.ToArray();
        }

        static ExprResultToken [] setVariable(Line line)
        {
            List<Line> result = new List<Line>();
            List<Line> rawresult = new List<Line>();
            List<ExprToken> cline = convertToExprTree(line.tokenlist.GetRange(2, line.tokenlist.Count - 2));
            List<ExprResultToken> init;
            if (cline.Count > 1)
            {
                init = exprBreakdown(ref cline);
                try
                {
                    ExprResultToken lastLine = init[init.Count - 1];
                    ExprToken finalReference = lastLine.container[lastLine.container.Count - 1];
                    ExprToken declareToken = new ExprToken("raw", "asg");
                    ExprToken variableName = new ExprToken("raw", line.tokenlist[0].value);
                    ExprResultToken declaration = new ExprResultToken(new ExprToken[] { declareToken, variableName, finalReference });
                    init.Add(declaration);
                    /*foreach (ExprResultToken l in init)
                    {
                        Console.WriteLine(l);
                    }*/
                }
                catch (Exception ex)
                {

                }
            }
            else
            {
                init = new List<ExprResultToken>();
                ExprToken declareToken = new ExprToken("raw", "dcl");
                ExprToken variableName = new ExprToken("raw", line.tokenlist[0].value);
                ExprToken finalReference = new ExprToken("raw", line.tokenlist[2].value);
                ExprResultToken declaration = new ExprResultToken(new ExprToken[] { declareToken, variableName, finalReference });
                init.Add(declaration);
                /*foreach (ExprResultToken l in init)
                {
                    Console.WriteLine(l);
                }*/
            }
            return init.ToArray();
        }

        static ExprResultToken [] startLoop(Line line)
        {
            List<ExprToken> cline = convertToExprTree(line.tokenlist);

            ExprToken labelToken = new ExprToken("raw", "loop" + loopCount + ":");
            //
            ExprToken cmpToken = new ExprToken("raw", "cmp");
            ExprToken varCond = cline[2];
            ExprToken comparitor = new ExprToken("raw", ">=");
            ExprToken varEndCond = cline[4];
            ExprToken trueToken = new ExprToken("raw", "lt" + loopCount);
            ExprToken falseToken = new ExprToken("raw", "lf" + loopCount);
            //
            ExprToken trueTokenLabel = new ExprToken("raw", "lt" + loopCount + ":");
            ExprToken jmpToken = new ExprToken("raw", "jmp");
            ExprToken endToken = new ExprToken("raw", "lend" + loopCount);
            //
            ExprToken falseTokenLabel = new ExprToken("raw", "lf" + loopCount + ":");
            //
            ExprResultToken label = new ExprResultToken(new ExprToken[] { labelToken });
            ExprResultToken condition = new ExprResultToken(new ExprToken[] {cmpToken, varCond, comparitor, varEndCond, trueToken, falseToken});
            ExprResultToken trueLine = new ExprResultToken(new ExprToken[] {trueTokenLabel});
            ExprResultToken endLine = new ExprResultToken(new ExprToken[] {jmpToken, endToken});
            ExprResultToken falseLine = new ExprResultToken(new ExprToken[] { falseTokenLabel });
            return new ExprResultToken[] {label, condition, trueLine, endLine, falseLine};
        }

        static ExprResultToken[] endLoop(Line line)
        {
            ExprToken jmpToken = new ExprToken("raw", "jmp");
            ExprToken loopToken = new ExprToken("raw", "loop" + loopCount);
            ExprToken endLabel = new ExprToken("raw", "lend" + loopCount + ":");
            ExprResultToken jmpLine = new ExprResultToken(new ExprToken[] { jmpToken, loopToken});
            ExprResultToken endLine = new ExprResultToken(new ExprToken[] { endLabel });
            return new ExprResultToken[] { jmpLine, endLine };
        }

        static ExprResultToken[] execute(Line line)
        {
            if (line.tokenlist[1] == "forward")
            {
                ExprToken fwdToken = new ExprToken("raw", "fwd");
                ExprResultToken fwd = new ExprResultToken(new ExprToken[] { fwdToken });
                return new ExprResultToken[] { fwd };
            }
            else
            {
                throw new Exception("Cannot find command");
            }
        }

        static ExprResultToken[] turn(Line line)
        {
            if (line.tokenlist[1] == "left")
            {
                ExprToken lftToken = new ExprToken("raw", "lft");
                ExprResultToken lft = new ExprResultToken(new ExprToken[] { lftToken });
                return new ExprResultToken[] { lft };
            }
            else if (line.tokenlist[1] == "right")
            {
                ExprToken rgtToken = new ExprToken("raw", "rgt");
                ExprResultToken rgt = new ExprResultToken(new ExprToken[] { rgtToken });
                return new ExprResultToken[] { rgt };
            }
            else
            {
                throw new Exception("Cannot find command");
            }
        }

        static ExprResultToken[] sense(Line line)
        {
            List<ExprToken> cline = convertToExprTree(line.tokenlist);
            ExprToken senseToken = new ExprToken("raw", "sns");
            if (!variableExist(cline[2].ToString()))
            {
                throw new Exception("Variable " + cline[2] + " not found");
            }
            ExprResultToken sense = new ExprResultToken(new ExprToken[] { senseToken, cline[2] });
            return new ExprResultToken[] { sense };

        }

        static ExprResultToken[] ifCMD(Line line)
        {
            List<ExprToken> cline = convertToExprTree(line.tokenlist);

            ExprToken cmpToken = new ExprToken("raw", "cmp");
            ExprToken varCond = cline[2];
            ExprToken comparitor = new ExprToken("raw", ">=");
            ExprToken varEndCond = cline[4];
            ExprToken trueToken = new ExprToken("raw", "t" + ifCount);
            ExprToken falseToken = new ExprToken("raw", "f" + ifCount);
            //
            ExprToken trueTokenLabel = new ExprToken("raw", "t" + ifCount + ":");

            //
            ExprResultToken condition = new ExprResultToken(new ExprToken[] { cmpToken, varCond, comparitor, varEndCond, trueToken, falseToken });
            ExprResultToken trueLine = new ExprResultToken(new ExprToken[] { trueTokenLabel });
            return new ExprResultToken[] { condition, trueLine };
        }

        static ExprResultToken[] elseCMD(Line line)
        {
            ExprToken jmpToken = new ExprToken("raw", "jmp");
            ExprToken endToken = new ExprToken("raw", "end" + ifCount);
            //
            ExprToken falseTokenLabel = new ExprToken("raw", "f" + ifCount + ":");
            //
            ExprResultToken endLine = new ExprResultToken(new ExprToken[] { jmpToken, endToken });
            ExprResultToken falseLine = new ExprResultToken(new ExprToken[] { falseTokenLabel });
            return new ExprResultToken[] { endLine, falseLine };
        }

        static ExprResultToken[] endIfCMD(Line line)
        {
            ExprToken jmpToken = new ExprToken("raw", "jmp");
            ExprToken endToken = new ExprToken("raw", "end" + ifCount);

            ExprToken endTokenLabel = new ExprToken("raw", "end" + ifCount + ":");

            ExprResultToken endLine = new ExprResultToken(new ExprToken[] { jmpToken, endToken });
            ExprResultToken end = new ExprResultToken(new ExprToken[] { endTokenLabel });

            return new ExprResultToken[] { endLine, end };
        }

        static List<ExprToken> convertToExprTree(List<Token> tokens)
        {
            List<ExprToken> result = new List<ExprToken>();
            for (int i = 0; i < tokens.Count; i++)
            {
                result.Add(new ExprToken("raw", tokens[i].value));
            }
            return result;
        }

        static List<ExprResultToken> exprBreakdown(ref List<ExprToken> line)
        {
            List<ExprResultToken> result = new List<ExprResultToken>();
            Queue<ExprResultToken> pStack = new Queue<ExprResultToken>();
            Stack<ExprToken> aStack = new Stack<ExprToken>();
            foreach (ExprToken token in line)
            {
                if (token.getType() == "raw")
                {

                    if (token == ")")
                    {
                        List<ExprToken> tokenList = new List<ExprToken>();
                        ExprToken cToken = aStack.Pop();
                        while (cToken != "(")
                        {
                            tokenList.Add(cToken);
                            cToken = aStack.Pop();
                        }
                        tokenList.Reverse();
                        ExprResultToken popResult = new ExprResultToken(tokenList.ToArray());
                        pStack.Enqueue(popResult);
                        aStack.Push(new ExprToken("@" + (pStack.Count).ToString(), popResult));
                    }
                    else
                    {
                        if (token !="(" && token != "+" && token != "-" && token != "*" && token != "/" && token.ToString()[0] != '@' && !isNumber(token.ToString()))
                        {
                            if (!variableExist(token.ToString()))
                            {
                                throw new Exception("No variable found error occured for " + token);
                            }
                        }
                        aStack.Push(token);
                    }
                }
            }

            List<ExprToken> tokenList2 = new List<ExprToken>();
            while (aStack.Count != 0)
            {
                ExprToken cToken = aStack.Pop();
                tokenList2.Add(cToken);
            }
            tokenList2.Reverse();
            ExprResultToken popResult2 = new ExprResultToken(tokenList2.ToArray());
            pStack.Enqueue(popResult2);
            int reductionCount = 0;
            foreach(ExprResultToken reduction in pStack)
            {
                bool done = false;
                while(!done)
                {
                    bool operation = false;
                    for (int i = 0; i < reduction.container.Count; i++)
                    {
                        ExprToken token = reduction.container[i];
                        if (token == "*")
                        {
                            ExprToken op1 = reduction.container[i - 1];
                            ExprToken op2 = reduction.container[i + 1];
                            reduction.container.RemoveAt(i + 1);
                            reduction.container.RemoveAt(i);
                            reduction.container[i - 1] = new ExprToken("raw", "@" + (reductionCount + 1).ToString());
                            operation = true;
                            ExprToken mul = new ExprToken("raw", "mul");
                            ExprToken tmp = new ExprToken("raw", "@" + (reductionCount + 1).ToString());
                            ExprResultToken multiply = new ExprResultToken(new ExprToken[]{ mul, op1, op2, tmp});
                            result.Add(multiply);
                            i = 0;
                        }
                    }

                    for (int i = 0; i < reduction.container.Count; i++)
                    {
                        ExprToken token = reduction.container[i];
                        if (token == "+")
                        {
                            ExprToken op1 = reduction.container[i - 1];
                            ExprToken op2 = reduction.container[i + 1];
                            reduction.container.RemoveAt(i + 1);
                            reduction.container.RemoveAt(i);
                            reduction.container[i - 1] = new ExprToken("raw", "@" + (reductionCount + 1).ToString());
                            operation = true;
                            ExprToken add = new ExprToken("raw", "add");
                            ExprToken tmp = new ExprToken("raw", "@" + (reductionCount + 1).ToString());
                            ExprResultToken multiply = new ExprResultToken(new ExprToken[] { add, op1, op2, tmp });
                            result.Add(multiply);
                            i = 0;
                        }
                        else if (token == "-")
                        {
                            ExprToken op1 = reduction.container[i - 1];
                            ExprToken op2 = reduction.container[i + 1];
                            reduction.container.RemoveAt(i + 1);
                            reduction.container.RemoveAt(i);
                            reduction.container[i - 1] = new ExprToken("raw", "@" + (reductionCount + 1).ToString());
                            operation = true;
                            ExprToken sub = new ExprToken("raw", "sub");
                            ExprToken tmp = new ExprToken("raw", "@" + (reductionCount + 1).ToString());
                            ExprResultToken multiply = new ExprResultToken(new ExprToken[] { sub, op1, op2, tmp });
                            result.Add(multiply);
                            i = 0;
                        }
                    }
                    done = !operation;
                }
                reductionCount += 1;
            }

            return result;

        }
    }
}
